from pandas import read_csv
import pickle
from math import ceil
from more_itertools import grouper
from time import time
from fw_ddsm.parameter import *


def new_pricing_table(normalised_pricing_table_csv, maximum_demand_level,
                      num_periods=no_intervals, weight=pricing_table_weight):
    # ---------------------------------------------------------------------- #
    # normalised_pricing_table_csv:
    #       the path of the CSV file of the normalised pricing table
    # demand_level_scalar:
    #       the scalar for rescaling the normalised demand levels
    # ---------------------------------------------------------------------- #

    csv_table = read_csv(normalised_pricing_table_csv, header=None)
    num_levels = len(csv_table.index)
    demand_level_scalar = maximum_demand_level * weight
    csv_table.loc[num_levels + 1] = [csv_table[0].values[-1] * 10] + [demand_level_scalar * 1.2 for _ in
                                                                      range(num_periods)]

    zero_digit = 100
    pricing_table = dict()
    pricing_table[p_price_levels] = list(csv_table[0].values)
    pricing_table[p_demand_table] = dict()
    pricing_table[p_demand_table] = \
        {period:
             {level:
                  ceil(csv_table[period + 1].values[level] * demand_level_scalar / zero_digit) * zero_digit
              for level in range(len(csv_table[period + 1]))}
         for period in range(num_periods)}

    return pricing_table



from pandas import DataFrame as df
from fw_ddsm.scripts.custom_functions import *
from fw_ddsm.parameter import *


class Tracker:

    def __init__(self, ):
        self.data = dict()
        self.name = "name"

    def new(self, name=""):
        self.name = name

        for key in [s_starts, s_demand, s_demand_max, s_demand_total, s_demand_reduction, s_par,
                    s_penalty, s_obj, p_prices, p_cost, p_cost_reduction, p_step, t_time, b_profile, s_debugger]:
            self.data[key] = dict()

    def read(self, existing_tracker):
        self.data = dict()
        self.data = existing_tracker.copy()

    def update(self, num_record, tracker_data=None, demands=None, prices=None, penalty=None,
               run_time=None, cost=None, step=None, init_demand_max=None, init_cost=None, tasks_starts=None,
               battery_profile=None, obj_par=False, debugger=None):
        obj = 0
        if tracker_data is None:
            tracker_data = self.data
        if step is not None:
            tracker_data[p_step][num_record] = round(step, 4)
        if prices is not None:
            tracker_data[p_prices][num_record] = prices
        if cost is not None:
            cost2 = round(cost, 2)
            tracker_data[p_cost][num_record] = cost2
            obj += cost2
            if init_cost is not None:
                tracker_data[p_cost_reduction][num_record] = round((init_cost - cost) / init_cost, 2)
        if penalty is not None:
            penalty2 = round(penalty, 2)
            tracker_data[s_penalty][num_record] = penalty2
            obj += penalty2
        if demands is not None:
            demand_max = round(max(demands), 2)
            tracker_data[s_demand][num_record] = demands
            tracker_data[s_demand_max][num_record] = demand_max
            tracker_data[s_demand_total][num_record] = round(sum(demands), 2)
            par = round(demand_max / average(demands), 2)
            tracker_data[s_par][num_record] = par
            obj += int(obj_par) * par
            if init_demand_max is not None:
                tracker_data[s_demand_reduction][num_record] \
                    = round((init_demand_max - demand_max) / init_demand_max, 2)
        if run_time is not None:
            tracker_data[t_time][num_record] = round(run_time, 4)
        if tasks_starts is not None:
            tracker_data[s_starts][num_record] = tasks_starts
        if battery_profile is not None:
            tracker_data[b_profile][num_record] = battery_profile
        if debugger is not None:
            tracker_data[s_debugger][num_record] = debugger

        tracker_data[s_obj][num_record] = obj

        return tracker_data

    def extract_data(self):

        demands = self.data[s_demand]
        batteries = self.data[b_profile]
        prices = self.data[p_prices]
        others = {k: self.data[k]
                  for k in [s_par, s_demand_reduction, p_cost_reduction, s_penalty,
                            s_demand_total, s_demand_max, p_cost, s_obj, t_time, p_step]}
        debugger = self.data[s_debugger]
        return demands, batteries, prices, others, debugger

    def write_to_file(self, folder, print_demands=True, print_prices=True, print_others=True,
                      print_batteries=True, print_debugger=True):
        demands, batteries, prices, others, debugger = self.extract_data()
        if print_demands:
            df_demands = df.from_dict(demands).div(1000)
            df_demands.to_csv(r"{}{}_demands.csv".format(folder, self.name))
        if print_prices:
            df_prices = df.from_dict(prices)
            df_prices.to_csv(r"{}{}_prices.csv".format(folder, self.name))
        if print_others:
            df_others = df.from_dict(others)
            df_others.to_csv(r"{}{}_others.csv".format(folder, self.name))
        if print_batteries:
            df_batteries = df.from_dict(batteries)
            df_batteries.to_csv(r"{}{}_batteries.csv".format(folder, self.name))
        if print_debugger:
            df_debugger = df.from_dict(debugger)
            df_debugger.to_csv(r"{}{}_debugger.csv".format(folder, self.name))




import random as r
from numpy import sqrt, pi, random
from fw_ddsm.parameter import *


def new_task(
        mode_value, list_of_devices_power, pst_probabilities, max_care_factor, scheduling_window_width,
        num_intervals=no_intervals, num_periods=no_periods, num_intervals_periods=no_intervals_periods
):
    # ---------------------------------------------------------------------- #
    # mode_value:
    #       a parameter for generation the duration using the Rayleigh distribution
    # list_of_device_power:
    #       the list of commonly used devices' power, or the list of devices' power rates
    # pst_probabilities:
    #       the probability distribution used for sampling the PST
    # max_care_factor:
    #       the maximum care factor, e.g. 1, or 10 or 100
    # scheduling_window_width:
    #       "full", can be rescheduled to any time intervals;
    #       "fixed", cannot be moved;
    # ---------------------------------------------------------------------- #

    # task power
    power = r.choice(list_of_devices_power)
    power = int(power * 1000)

    # task duration
    duration = max(1, int(random.rayleigh(mode_value, 1)[0]))

    # task preferred start time
    preferred_start_time = max(int(random.choice(a=num_periods, size=1, p=pst_probabilities)[0]) *
                               num_intervals_periods + r.randint(-num_intervals_periods + 1, num_intervals_periods),
                               0)
    preferred_start_time = min(preferred_start_time, num_intervals - 1)

    # task earliest starting time and latest finish time
    if scheduling_window_width == "full":
        earliest_start_time = 0
        latest_finish_time = num_intervals - 1 + duration
    elif scheduling_window_width == "fixed":
        earliest_start_time = preferred_start_time
        latest_finish_time = preferred_start_time + duration - 1
    else:
        earliest_start_time = r.randint(0, preferred_start_time)
        latest_finish_time = r.randint(preferred_start_time + duration - 1, num_intervals - 1 + duration)

    # task care factor
    care_factor = int(r.choice([i for i in range(1, max_care_factor + 1)]))

    return power, duration, preferred_start_time, earliest_start_time, latest_finish_time, care_factor


def new_dependent_tasks(num_intervals, num_tasks_dependent, num_total_tasks,
                        preferred_starts, durations, earliest_starts, latest_ends, ensure_dependent=False):
    precedors = dict()
    no_precedences = 0
    succ_delays = dict()

    def retrieve_precedes(list0):
        list3 = []
        for l in list0:
            if l in precedors:
                list2 = precedors[l]
                retrieved_list = retrieve_precedes(list2)
                list3.extend(retrieved_list)
            else:
                list3.append(l)
        return list3

    def add_precedes(task, previous, delay):
        if task not in precedors:
            precedors[task] = [previous]
            succ_delays[task] = [delay]
        else:
            precedors[task].append(previous)
            succ_delays[task].append(delay)

    for t in range(num_total_tasks - num_tasks_dependent, num_total_tasks):
        if r.choice([True, ensure_dependent]):
            previous_tasks = list(range(t))
            r.shuffle(previous_tasks)
            for prev in previous_tasks:
                if preferred_starts[prev] + durations[prev] - 1 < preferred_starts[t] \
                        and earliest_starts[prev] + durations[prev] < latest_ends[t] - durations[t] + 1:

                    if prev not in precedors:
                        # feasible delay
                        succeding_delay = num_intervals - 1
                        add_precedes(t, prev, succeding_delay)
                        no_precedences += 1
                        break
                    else:
                        # find all precedors of this previous task
                        precs_prev = retrieve_precedes([prev])
                        precs_prev.append(prev)

                        precs_prev_duration = sum([durations[x] for x in precs_prev])
                        latest_pstart = preferred_starts[precs_prev[0]]
                        latest_estart = earliest_starts[precs_prev[0]]

                        if latest_pstart + precs_prev_duration - 1 < preferred_starts[t] \
                                and latest_estart + precs_prev_duration < latest_ends[t] - durations[t] + 1:
                            succeding_delay = num_intervals - 1
                            add_precedes(t, prev, succeding_delay)
                            no_precedences += 1
                            break

    return no_precedences, precedors, succ_delays


def new_household(
        preferred_demand_profile, list_of_devices_power,
        num_intervals=no_intervals, num_periods=no_periods, num_intervals_periods=no_intervals_periods,
        max_demand_multiplier=maximum_demand_multiplier,
        num_tasks_dependent=no_tasks_dependent, ensure_dependent=False,
        full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
        semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
        fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
        inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
        household_id=0,
        capacity_max=battery_capacity_max, capacity_min=battery_capacity_min,
        power=battery_power, efficiency=battery_efficiency
):

    pst_probabilities = [int(p) for p in preferred_demand_profile]
    sum_pst_probabilities = sum(pst_probabilities)
    pst_probabilities = [p / sum_pst_probabilities for p in pst_probabilities]

    # I meant mean value is 40 minutes
    mean_value = 40.0 / (24.0 * 60.0 / num_intervals)
    mode_value = sqrt(2 / pi) * mean_value

    # task details
    preferred_starts = []
    earliest_starts = []
    latest_ends = []
    durations = []
    powers = []
    care_factors = []
    household_demand_profile = [0] * num_intervals

    # new tasks in a household
    def get_new_tasks(num_tasks, scheduling_window_width):
        for counter_j in range(num_tasks):
            demand, duration, p_start, e_start, l_finish, care_f \
                = new_task(mode_value, list_of_devices_power, pst_probabilities,
                           max_care_factor, scheduling_window_width,
                           num_intervals=num_intervals, num_periods=num_periods,
                           num_intervals_periods=num_intervals_periods)

            powers.append(demand)
            durations.append(duration)
            preferred_starts.append(p_start)
            earliest_starts.append(e_start)
            latest_ends.append(l_finish)
            care_factors.append(care_f)
            # add this task demand to the household demand
            for d in range(duration):
                household_demand_profile[(p_start + d) % num_intervals] += demand

    fixed_task_max = max(fixed_task_max, fixed_task_min)
    semi_flex_task_max = max(semi_flex_task_max, semi_flex_task_min)
    full_flex_task_max = max(full_flex_task_max, full_flex_task_min)

    num_fixed_tasks = r.randint(fixed_task_min, fixed_task_max)
    num_semi_flex_tasks = r.randint(semi_flex_task_min, semi_flex_task_max)
    num_full_flex_tasks = r.randint(full_flex_task_min, full_flex_task_max)

    get_new_tasks(num_fixed_tasks, "fixed")
    get_new_tasks(num_semi_flex_tasks, "semi")
    get_new_tasks(num_full_flex_tasks, "full")

    # set the household demand limit
    maximum_demand = sum(powers) * max_demand_multiplier

    # precedence among tasks
    num_total_tasks = num_full_flex_tasks + num_semi_flex_tasks + num_fixed_tasks
    no_precedences, precedors, succ_delays \
        = new_dependent_tasks(num_intervals, num_tasks_dependent, num_total_tasks,
                              preferred_starts, durations, earliest_starts, latest_ends, ensure_dependent)

    household = dict()
    if household_id is not None:
        household[h_key] = household_id

    # save the details of new tasks
    household[h_psts] = preferred_starts
    household[h_ests] = earliest_starts
    household[h_lfs] = latest_ends
    household[h_durs] = durations
    household[h_powers] = powers
    household[h_cfs] = care_factors
    household[h_max_cf] = max_care_factor
    household[h_no_precs] = no_precedences
    household[h_precs] = precedors
    household[h_succ_delay] = succ_delays
    household[h_demand_limit] = maximum_demand
    household[h_incon_weight] = inconvenience_cost_weight
    household[s_demand] = household_demand_profile

    # save the battery details
    household[b_cap_max] = capacity_max
    household[b_cap_min] = capacity_min
    household[b_power] = power
    household[b_profile] = [0] * num_intervals
    household[b_eff] = efficiency

    # return the household details
    return household



from fw_ddsm.community import *
from fw_ddsm.aggregator import *


class Iteration:

    def __init__(self):
        self.tasks_scheduling_method = ""
        self.pricing_method = ""
        self.num_households = 0
        self.num_iteration = 0
        self.num_intervals = 0

        self.community = Community()
        self.aggregator = Aggregator()
        self.data_folder = "data/"
        self.start_time_probability = [1] * no_periods

    def new(self, algorithm, num_households,
            num_intervals=no_intervals,
            file_task_power=file_demand_list, max_demand_multiplier=maximum_demand_multiplier,
            file_normalised_pricing_table=file_pricing_table, file_preferred_demand_profile=file_pdp,
            num_tasks_dependent=no_tasks_dependent, ensure_dependent=False,
            full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
            semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
            fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
            inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
            data_folder=None, backup_data_folder=None,
            date_time=None,
            capacity_max=battery_capacity_max, capacity_min=battery_capacity_min,
            power=battery_power, efficiency=battery_efficiency):

        if data_folder is not None:
            if not data_folder.endswith("/"):
                data_folder += "/"
            self.data_folder = data_folder
        else:
            data_folder = self.data_folder

        self.tasks_scheduling_method = algorithm[m_before_fw]
        self.pricing_method = algorithm[m_after_fw]
        self.num_households = num_households

        # 1. generate new households, trackers and a pricing table
        preferred_demand_profile \
            = self.community.new(file_preferred_demand_profile=file_preferred_demand_profile,
                                 file_demand_list=file_task_power,
                                 tasks_scheduling_method=self.tasks_scheduling_method,
                                 num_intervals=num_intervals,
                                 num_households=self.num_households,
                                 max_demand_multiplier=max_demand_multiplier,
                                 num_tasks_dependent=num_tasks_dependent, ensure_dependent=ensure_dependent,
                                 full_flex_task_min=full_flex_task_min, full_flex_task_max=full_flex_task_max,
                                 semi_flex_task_min=semi_flex_task_min, semi_flex_task_max=semi_flex_task_max,
                                 fixed_task_min=fixed_task_min, fixed_task_max=fixed_task_max,
                                 inconvenience_cost_weight=inconvenience_cost_weight, max_care_factor=max_care_factor,
                                 write_to_file_path=data_folder, backup_file_path=backup_data_folder,
                                 date_time=date_time,
                                 capacity_max=capacity_max, capacity_min=capacity_min,
                                 power=power, efficiency=efficiency
                                 )

        prices, preferred_cost \
            = self.aggregator.new_aggregator(normalised_pricing_table_csv=file_normalised_pricing_table,
                                             aggregate_preferred_demand_profile=preferred_demand_profile,
                                             pricing_method=self.pricing_method,
                                             write_to_file_path=data_folder, backup_file_path=backup_data_folder,
                                             date_time=date_time)

        return preferred_demand_profile, prices

    def read(self, algorithm, inconvenience_cost_weight=None, new_dependent_tasks=None, ensure_dependent=False,
             capacity_max=battery_capacity_max, capacity_min=battery_capacity_min, power=battery_power,
             efficiency=battery_efficiency,
             read_from_folder="data/", date_time=None):

        if read_from_folder is None:
            read_from_folder = self.data_folder

        self.tasks_scheduling_method = algorithm[m_before_fw]
        self.pricing_method = algorithm[m_after_fw]
        preferred_demand_profile \
            = self.community.read(read_from_folder=read_from_folder,
                                  tasks_scheduling_method=self.tasks_scheduling_method,
                                  inconvenience_cost_weight=inconvenience_cost_weight,
                                  capacity_max=capacity_max, capacity_min=capacity_min, power=power,
                                  efficiency=efficiency,
                                  num_dependent_tasks=new_dependent_tasks, ensure_dependent=ensure_dependent,
                                  date_time=date_time)
        prices, preferred_cost \
            = self.aggregator.read_aggregator(read_from_folder=read_from_folder, date_time=date_time,
                                              pricing_method=self.pricing_method,
                                              aggregate_preferred_demand_profile=preferred_demand_profile)

        return preferred_demand_profile, prices

    def begin_iteration(self, starting_prices,
                        use_battery=False, battery_model=None, battery_solver=None,
                        num_cpus=None, timeout=time_out, fully_charge_time=fully_charge_hour,
                        min_step_size=min_step, roundup_tiny_step=False,
                        print_done=False, print_steps=False):

        scheduling_method = self.tasks_scheduling_method
        pricing_method = self.pricing_method
        prices = starting_prices

        if use_battery:
            battery_model = file_mip_battery if battery_model is None else battery_model
            battery_solver = "mip" if battery_solver is None else battery_solver

        num_iteration = 1
        step = 0.9
        obj_pre = 0
        obj_fw = -1
        obj_improve = 1
        while step > 0.0005 and obj_improve > 0.001:
            aggregate_demand_profile, aggregate_battery_profile, \
            weighted_total_inconvenience, time_scheduling_iteration, total_obj \
                = self.community.schedule(num_iteration=num_iteration, prices=prices,
                                          pricing_table=self.aggregator.pricing_table,
                                          tasks_scheduling_method=scheduling_method,
                                          use_battery=use_battery,
                                          battery_model=battery_model, battery_solver=battery_solver,
                                          num_cpus=num_cpus, timeout=timeout,
                                          fully_charge_time=fully_charge_time,
                                          print_upon_completion=print_done)

            prices, consumption_cost, inconvenience, step, \
            new_aggregate_demand_profile, new_aggregate_battery_profile, time_pricing \
                = self.aggregator.pricing(num_iteration=num_iteration,
                                          aggregate_demand_profile=aggregate_demand_profile,
                                          aggregate_battery_profile=aggregate_battery_profile,
                                          total_obj=total_obj,
                                          aggregate_inconvenience=weighted_total_inconvenience,
                                          min_step_size=min_step_size,
                                          roundup_tiny_step=roundup_tiny_step, print_steps=print_steps)

            if num_iteration > 1:
                obj_pre = obj_fw
            obj_fw = consumption_cost + inconvenience
            obj_improve = obj_pre - obj_fw
            num_iteration += 1

        print(f"Converged in {num_iteration - 1}")

        self.start_time_probability = self.aggregator.compute_start_time_probabilities()
        return self.start_time_probability, num_iteration - 1

    def finalise_schedules(self, start_time_probability=None, tasks_scheduling_method=None, num_samples=1):
        if tasks_scheduling_method is None:
            tasks_scheduling_method = self.tasks_scheduling_method
        if start_time_probability is None:
            start_time_probability = self.start_time_probability
        for i in range(1, num_samples + 1):
            final_aggregate_demand_profile, final_battery_profile, final_total_inconvenience \
                = self.community.finalise_schedule(num_sample=i,
                                                   tasks_scheduling_method=tasks_scheduling_method,
                                                   start_probability_distribution=start_time_probability)
            prices, consumption_cost, inconvenience, step, \
            new_aggregate_demand_profile, new_aggregate_battery_profile, time_pricing \
                = self.aggregator.pricing(num_iteration=i,
                                          aggregate_demand_profile=final_aggregate_demand_profile,
                                          aggregate_battery_profile=final_battery_profile,
                                          aggregate_inconvenience=final_total_inconvenience,
                                          finalising=True)
        # return consumption_cost, inconvenience
